import json
from pathlib import Path

from pymultirole_plugins.v1.schema import Document, DocumentList

from pyprocessors_pseudonimizer.pseudonimizer import (
    PseudonimizerProcessor,
    PseudonimizerParameters, OPERATOR_IDENTITY_STR
)


def test_normalize():
    testdir = Path(__file__).parent
    source = Path(
        testdir,
        "data/afp_ner_fr-document-test.json",
    )
    with source.open("r") as fin:
        jdoc = json.load(fin)
        doc = Document(**jdoc)
        formatter = PseudonimizerProcessor()
        options = PseudonimizerParameters()
        options.mapping = {
            'wikidata': OPERATOR_IDENTITY_STR,
            'organization': json.dumps({
                "type": "mask",
                "masking_char": "X",
                "chars_to_mask": 3,
                "from_end": False,
            }),
            'afporganization': json.dumps({
                "type": "faker",
                "provider": "company",
                "locale": "en_US"
            }),
            'person': json.dumps({
                "type": "faker",
                "provider": "name",
                "locale": "fr_FR"
            }),
            'afpperson': json.dumps({
                "type": "faker",
                "provider": "name",
                "locale": "fr_FR"
            }
            )
        }
        anonymizeds = formatter.process([doc], options)
        norm_file = testdir / "data/afp_ner_fr-document-test-anon.json"
        dl = DocumentList(__root__=anonymizeds)
        with norm_file.open("w") as fout:
            print(dl.json(exclude_none=True, exclude_unset=True, indent=2), file=fout)



# Copyright: See the LICENSE file.


"""Additional declarations for "faker" attributes.

Usage:

    class MyFactory(factory.Factory):
        class Meta:
            model = MyProfile

        first_name = factory.Faker('name')
"""

import contextlib

import faker
import faker.config


class FakerOp:
    """Wrapper for 'faker' values.

    Args:
        provider (str): the name of the Faker field
        locale (str): the locale to use for the faker

        All other kwargs will be passed to the underlying provider
        (e.g ``factory.Faker('ean', length=10)``
        calls ``faker.Faker.ean(length=10)``)

    Usage:
        >>> foo = FakerOp('name')
    """

    def __init__(self, provider, **kwargs):
        self.locale = kwargs.pop('locale', None)
        self.provider = provider
        self.extra = kwargs

    def evaluate(self, val):
        subfaker = self._get_faker(self.locale)
        result = subfaker.format(self.provider, **self.extra)
        return result

    _FAKER_REGISTRY = {}
    _DEFAULT_LOCALE = faker.config.DEFAULT_LOCALE

    @classmethod
    @contextlib.contextmanager
    def override_default_locale(cls, locale):
        old_locale = cls._DEFAULT_LOCALE
        cls._DEFAULT_LOCALE = locale
        try:
            yield
        finally:
            cls._DEFAULT_LOCALE = old_locale

    @classmethod
    def _get_faker(cls, locale=None):
        if locale is None:
            locale = cls._DEFAULT_LOCALE

        if locale not in cls._FAKER_REGISTRY:
            subfaker = faker.Faker(locale=locale)
            cls._FAKER_REGISTRY[locale] = subfaker

        return cls._FAKER_REGISTRY[locale]

    @classmethod
    def add_provider(cls, provider, locale=None):
        """Add a new Faker provider for the specified locale"""
        cls._get_faker(locale).add_provider(provider)



import json
from functools import lru_cache
from typing import List, Type, cast, Dict

from presidio_anonymizer import AnonymizerEngine
from presidio_anonymizer.entities import RecognizerResult, OperatorConfig
from pydantic import Field, BaseModel
from pymultirole_plugins.v1.processor import ProcessorParameters, ProcessorBase
from pymultirole_plugins.v1.schema import Document, AltText

from .fakerop import FakerOp

OPERATOR_EXAMPLE1 = {
    "type": "mask",
    "masking_char": "*",
    "chars_to_mask": 12,
    "from_end": True,
}
OPERATOR_EXAMPLE1_STR = json.dumps(OPERATOR_EXAMPLE1, indent=2)

OPERATOR_EXAMPLE2 = {
    "type": "replace",
    "new_value": "_ANONYMIZED_"
}
OPERATOR_EXAMPLE2_STR = json.dumps(OPERATOR_EXAMPLE2, indent=2)

OPERATOR_LABEL = {
    "type": "label"
}
OPERATOR_LABEL_STR = json.dumps(OPERATOR_LABEL, indent=2)

OPERATOR_IDENTITY = {
    "type": "identity"
}
OPERATOR_IDENTITY_STR = json.dumps(OPERATOR_IDENTITY, indent=2)

CUSTOM_OPERATORS = {
    "identity": lambda x: x,
}

OPERATOR_FAKER = {
    "type": "faker",
    "provider": "name",
    "locale": "fr_FR"
}
OPERATOR_FAKER_STR = json.dumps(OPERATOR_FAKER, indent=2)


class PseudonimizerParameters(ProcessorParameters):
    mapping: Dict[str, str] = Field(None,
                                    description="List of anonymization/pseudonymization [operators](https://microsoft.github.io/presidio/tutorial/10_simple_anonymization/) like this ones:<br/><li>```" + OPERATOR_EXAMPLE1_STR + "```<li>```" + OPERATOR_EXAMPLE2_STR + "```<li>```" + OPERATOR_FAKER_STR + "```",
                                    extra="key:label,val:json")
    default_operator: str = Field(OPERATOR_LABEL_STR,
                                  description="""Default anonymization/pseudonymization operator to use if no explicit mapping is provided""",
                                  extra="json")
    as_altText: str = Field(
        "pseudonimization",
        description="""If defined generate the pseudonimization as an alternative text of the input document,
    if not replace the text of the input document.""",
    )


def operator_from_annotation(a, definitions):
    default_def = definitions.get(None, None)
    label = a.labelName or a.label
    if label in definitions:
        definition = definitions[label]
        return operator_from_definition(a, definition)
    elif default_def is not None:
        return operator_from_definition(a, default_def)
    return None


def operator_from_definition(a, definition):
    type = definition['type']
    if type == 'label':
        replacement = f"<{a.label or a.labelName}>"
        return OperatorConfig("replace",
                              {"new_value": replacement})
    elif type == "faker":
        definition.pop("type")
        provider = definition.pop("provider")
        fakerop = FakerOp(provider, **definition)
        return OperatorConfig("custom", {"lambda": fakerop.evaluate})
    elif type in CUSTOM_OPERATORS:
        return OperatorConfig("custom", {"lambda": CUSTOM_OPERATORS[type]})
    else:
        return OperatorConfig.from_json(definition)


class PseudonimizerProcessor(ProcessorBase):
    """Pseudonimizer."""

    def process(
            self, documents: List[Document], parameters: ProcessorParameters
    ) -> List[Document]:
        params: PseudonimizerParameters = cast(PseudonimizerParameters, parameters)
        mapping = frozenset(params.mapping.items()) if params.mapping else None
        definitions = get_definitions(mapping, params.default_operator)
        engine = get_engine()
        operators: Dict[str, OperatorConfig] = {}
        try:
            for document in documents:
                if document.annotations:
                    analyzer_results = []
                    for a in document.annotations:
                        if a.status != "KO":
                            analyzer_results.append(
                                RecognizerResult(entity_type=a.labelName, start=a.start, end=a.end, score=a.score))
                            if a.labelName not in operators:
                                op = operator_from_annotation(a, definitions)
                                if op is not None:
                                    operators[a.labelName] = op
                    result = engine.anonymize(
                        text=document.text,
                        analyzer_results=analyzer_results,
                        operators=operators,
                    )
                    if params.as_altText is not None and len(params.as_altText):
                        document.altTexts = document.altTexts or []
                        altTexts = [
                            alt
                            for alt in document.altTexts
                            if alt.name != params.as_altText
                        ]
                        altTexts.append(AltText(name=params.as_altText, text=result.text))
                        document.altTexts = altTexts
                    else:
                        document.text = result.text
                        document.annotations = None
                        document.sentences = None
        except BaseException as err:
            raise err
        return documents

    @classmethod
    def get_model(cls) -> Type[BaseModel]:
        return PseudonimizerParameters


@lru_cache(maxsize=None)
def get_engine():
    return AnonymizerEngine()


@lru_cache(maxsize=None)
def get_definitions(mapping_items, default):
    definitions = {None: None}
    if default is not None and default.strip() != "":
        default_def = json.loads(default)
        definitions[None] = default_def
    if mapping_items:
        for pname, pvalue in mapping_items:
            op_def = json.loads(pvalue)
            definitions[pname] = op_def
    return definitions
